/*
 *
 * $Id: k3bscsicommand_linux.cpp 402089 2005-03-31 09:01:07Z trueg $
 * Copyright (C) 2003 Sebastian Trueg <trueg@k3b.org>
 *
 * This file is part of the K3b project.
 * Copyright (C) 1998-2004 Sebastian Trueg <trueg@k3b.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */

#include "k3bscsicommand.h"
#include "k3bdevice.h"
#include "k3bscsicommand_solaris.h"
#include "k3bsolaris_uscsi.h"

#include <kdebug.h>

#include <sys/ioctl.h>
#include <sys/cdio.h>

#include <sys/dkio.h>
#include <sys/scsi/generic/commands.h>
#include <sys/scsi/impl/uscsi.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include <unistd.h>
#include <sys/types.h>

#if defined (__sun) || defined (SOLARIS)
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/scsi/scsi.h>
#include <errno.h>
extern int errno;
#endif

#define K3BERRCODE(s) ((((s)[2]&0x0F)<<16)|((s)[12]<<8)|((s)[13]))

extern "C" {
typedef struct solaris_uscsi
{
  struct uscsi_cmd ucmd;
  unsigned char* cdb; // [16]
  unsigned char* sense; // [18]
} solaris_uscsi;
}

class K3bDevice::ScsiCommand::Private
{
public:
  Private();
  ~Private();

public:
  int fd;
  solaris_uscsi u;
};

K3bDevice::ScsiCommand::Private::Private()
{
  fd = -1;
  u.cdb = (unsigned char *) ::malloc((size_t) 16);
  u.sense = (unsigned char *) ::malloc((size_t) 18);
  (void) memset (&u.ucmd, '\0', sizeof(struct uscsi_cmd));
  (void) memset(u.cdb, '\0', sizeof(u.cdb));
  (void) memset(u.sense, '\0', sizeof(u.sense));
}

K3bDevice::ScsiCommand::Private::~Private()
{
  ::free(u.cdb);
  ::free(u.sense);
}

void K3bDevice::ScsiCommand::clear()
{
  if (0L != d)
  {
    (void) memset (&d->u.ucmd, '\0', sizeof(struct uscsi_cmd));
    (void) memset(d->u.cdb, '\0', sizeof(d->u.cdb));
    (void) memset(d->u.sense, '\0', sizeof(d->u.sense));
  }
}


unsigned char& K3bDevice::ScsiCommand::operator[]( size_t i )
{
  if (0 == i)
  {
    (void) memset(&d->u.ucmd, '\0', sizeof(struct uscsi_cmd));
    (void) memset(d->u.cdb, '\0', size_t(16));
    (void) memset(d->u.sense, '\0', size_t(18));

    d->u.ucmd.uscsi_cdb = (caddr_t) d->u.cdb;
    d->u.ucmd.uscsi_rqbuf = (caddr_t) d->u.sense;
    d->u.ucmd.uscsi_rqlen = size_t(18); // sizeof(u.sense.)
    d->u.ucmd.uscsi_flags = (USCSI_SILENT  | USCSI_DIAGNOSE | USCSI_ISOLATE | USCSI_RQENABLE);
    // MINE:
    // d->u.ucmd.uscsi_flags = (USCSI_ISOLATE|USCSI_READ|USCSI_RQENABLE|USCSI_ASYNC);
    d->u.ucmd.uscsi_timeout = 1;
  }
  d->u.ucmd.uscsi_cdblen = i + 1;
  return d->u.cdb[i];
}


int K3bDevice::ScsiCommand::transport( TransportDirection dir,
				       void* data,
				       size_t len )
{
  int ret = 0;

  if (0 == solaris_uscsi_initialized())
    init_solaris_uscsi();

  d->u.ucmd.uscsi_cdb = (caddr_t) d->u.cdb;
  d->u.ucmd.uscsi_bufaddr = (caddr_t) data;
  d->u.ucmd.uscsi_buflen = len;
  d->u.ucmd.uscsi_flags |= dir;

  ret = send_solaris_uscsi (d->fd, &(d->u.ucmd));

  if (ret < 0)
    ret = ::ioctl(d->fd, USCSICMD, &d->u.ucmd);

  if (ret < 0)
  {
    if ((errno == EIO) && (d->u.sense[0] == '\0'))
    {
      size_t residue = d->u.ucmd.uscsi_resid;
      (void) memset(d->u.cdb, '\0', size_t(16));
      d->u.cdb[0] = 0x03;
      d->u.cdb[4] = size_t(18);
      d->u.ucmd.uscsi_cdblen = 6;
      d->u.ucmd.uscsi_bufaddr = (caddr_t) d->u.sense;
      d->u.ucmd.uscsi_buflen = size_t(18);
      d->u.ucmd.uscsi_flags = (USCSI_SILENT  | USCSI_DIAGNOSE | USCSI_ISOLATE | USCSI_READ);
      // MINE:
      // d->u.ucmd.uscsi_flags = (USCSI_ISOLATE|USCSI_READ|USCSI_RQENABLE|USCSI_ASYNC);
      // d->ucmd.uscsi_flags = (USCSI_SILENT|USCSI_DIAGNOSE|USCSI_ISOLATE|USCSI_READ);
      d->u.ucmd.uscsi_timeout = 1;
      ret = send_solaris_uscsi (d->fd, &(d->u.ucmd));
      d->u.ucmd.uscsi_resid = residue;
      if (ret < 0)
        return -1;
    }
    ret = K3BERRCODE(d->u.sense);
    ret = (ret < 0) ? -1 : 0;
  }
  return ret;
}

// vim: ts=2 sw=2 et

