#include <iostream>
using namespace std;

#include "k3bsolaris_scg.h"
#include "k3bsolaris_device.h"

#define USE_USCSI 1
#define SCSI_ATAPI 1
#define USE_SCGLIB 1

#define MAX_SCAN_DATA_LEN 30
#include "util.h"
#include "xconfig.h"
#include "standard.h"
#include "scg/scgcmd.h"
#include "scg/scsitransp.h"
#include "scg/scsireg.h"

#undef vendor
#undef product
#undef revision

#define MAX_SCG 16
#define MAX_TGT 16
#define MAX_LUN 8

struct scg_local
{
  union
  {
    int     SCG_files[MAX_SCG];
#ifdef  USE_USCSI
    short   scg_files[MAX_SCG][MAX_TGT][MAX_LUN];
#endif
   } u;
};

#define scglocal(p) ((struct scg_local *)((p)->local))
#define scgfiles(p) (scglocal(p)->u.SCG_files)

static int bus_atapi = 0;
static int kstat_init = false;

static int scanInquiry(SCSI *scgp, unsigned char *buf, SCGScanData* sdata, const char* scsiDevice);
void findSolarisScsiBusIdLun(const char* scsiDevice, SCGScanData* scanData)
{
  int b, l, t, len;
  int target = -1;
  unsigned char* buf = 0L;
  SCSI* scgp = 0L;
  char errstr[256];
  bool tryGuessLoop = true;

  const char* tgt;
  const char* ptr = ::strstr(scsiDevice, "/dev/rsr");
  if (0L != ptr)
  {
    tgt = &ptr[8];
    target = ::atoi(tgt);
  }
  else
  {
    ptr = ::strstr(scsiDevice, "/dev/rdsk/c");
    if (0L != ptr)
    {
      char c[2];
      tgt = &ptr[11];
      (void) sprintf(c, "%c", *tgt);
      target = ::atoi(c);
    }
  }

  scanData->bus = -1;
  scanData->target = target;
  scanData->lun = -1;
  scanData->is_valid = 0;

  len = 0;
  if ((scgp = scg_open(const_cast<char*>(scsiDevice), errstr, sizeof(errstr), 0, 0)) == NULL)
  {
    return;
  }

  buf = (unsigned char *) scg_getbuf(scgp, 100);
  if (buf == NULL)
  {
    return;
  }

  if (tryGuessLoop)
  {
    for (b = 0; b < MAX_SCG && len < MAX_SCAN_DATA_LEN; b++)
    {
      if (scg_havebus(scgp, b))
      {
        l = 0;
        for (t = 0; t < MAX_TGT && len < MAX_SCAN_DATA_LEN; t++)
        {
          scg_settarget(scgp, b, t, l);
          if (scanInquiry(scgp, buf, scanData, scsiDevice))
          {
            break;
          }
          else
            ++len;
        }
      }
    }
  }
  scg_close(scgp);
  if (target != -1)
    scanData->target = target;
}

int scanInquiry(SCSI *scgp, unsigned char *buf, SCGScanData* sdata, const char* scsiDevice)
{
  struct scg_cmd *cmd = scgp->scmd;
  int i;
  int is_atapi = 0;
  int cmd_status = 0;

  int dev_sense = scg_sense_key(scgp);

  memset(buf, '\0', 36);
  memset(cmd, '\0', sizeof(struct scg_cmd));

  cmd->cdb.g0_cdb.cmd = 0x12; // INQUIRY
  cmd->cdb.g0_cdb.count = 36;

  cmd->addr = (char*)buf;
  cmd->size = 36;
  cmd->flags = SCG_RECV_DATA|SCG_DISRE_ENA;
  cmd->cdb_len = SC_G0_CDBLEN;
  cmd->sense_len = CCS_SENSE_LEN;
  scgp->silent = 1;
  scgp->cmdname = "inquiry";

  is_atapi = scg_isatapi(scgp);
  cmd_status = scg_cmd(scgp);
  sdata->is_valid = 0;
  sdata->bus = -1;
  sdata->lun = -1;
  sdata->target = -1;

  if ((cmd_status == 0) || ((is_atapi == 1) && (dev_sense >= 0)))
  {
    struct scsi_inquiry* inq = (struct scsi_inquiry *) buf;
    if (inq->type == INQ_OPTD || inq->type == INQ_ROMD)
    {
      sdata->bus = scg_scsibus(scgp);
      sdata->target = scg_target(scgp);
      sdata->lun = scg_lun(scgp);

      strncpy(sdata->vendor, inq->vendor_info, 8);
      sdata->vendor[8] = '\0';
      strncpy(sdata->product, inq->prod_ident, 16);
      sdata->product[16] = '\0';
      strncpy(sdata->revision, inq->prod_revision, 4);
      sdata->revision[4] = '\0';
      sdata->is_valid = 1;
      return 1;
    }
  }
  return 0;
}


// vim: ts=2 sw=2 et

