#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <errno.h>
#include <priv.h>
#include <sys/smedia.h>

#include "k3bsolaris_uscsi.h"

extern int errno;
static priv_set_t* privset = 0L;
static void* smedia_lib = 0L;
static void* (*sm_get_handle)(int32_t);
static int (*sm_release_handle)(void*);
static int (*sm_uscsi_cmd)(void*, struct uscsi_cmd *);
static int initialized = 0;
static int priv_initialized = 0;

int solaris_priv_initialized(void)
{
  return priv_initialized;
}

int solaris_uscsi_initialized(void)
{
  return initialized;
}

void set_privileges(void)
{
  if (0L == privset)
    privset = priv_allocset();

  if (0L == privset)
  {
    perror("priv_allocset"), strerror(errno);
    return;
  }

  if (getppriv(PRIV_PERMITTED, privset) != 0)
  {
    perror("getppriv"), strerror(errno);
    return;
  }
  if (priv_addset(privset, "PRIV_SYS_DEVICES") != 0)
  {
    perror("priv_addset"), strerror(errno);
    return;
  }
  if (setppriv(PRIV_ON, PRIV_EFFECTIVE, privset) != 0)
  {
    perror("setppriv"), strerror(errno);
    return;
  }
  priv_initialized = 1;
}

void init_solaris_uscsi(void)
{
  if ((0L == smedia_lib) || (0L == sm_get_handle) || (0L == sm_release_handle) || (0L == sm_uscsi_cmd))
  {
    if (0L == smedia_lib)
      smedia_lib = dlopen("libsmedia.so.1", RTLD_NOW|RTLD_GLOBAL|RTLD_NODELETE);
    if (0L != smedia_lib)
    {
      if (0L == sm_get_handle)
        sm_get_handle = dlsym(smedia_lib, "smedia_get_handle");
      if (0L == sm_release_handle)
        sm_release_handle = dlsym(smedia_lib, "smedia_release_handle");
      if (0L == sm_uscsi_cmd)
        sm_uscsi_cmd = dlsym(smedia_lib, "smedia_uscsi_cmd");

      if ((0L == sm_get_handle) || (0L == sm_release_handle) || (0L == sm_uscsi_cmd))
      {
        sm_get_handle = 0L;
        sm_release_handle = 0L;
        sm_uscsi_cmd = 0L;
        dlclose(smedia_lib);
        smedia_lib = 0L;
        return;
      }
    }
  }
}


int send_solaris_uscsi(int fd, struct uscsi_cmd* psc)
{
  void *sm_hdl;
  int ret = -1;
  int32_t i;
  smwp_state_t sm_prot;

  init_solaris_uscsi();
  set_privileges();

  if (0L == sm_uscsi_cmd)
    sm_uscsi_cmd = dlsym(smedia_lib, "smedia_uscsi_cmd");

  if (sm_uscsi_cmd && (sm_hdl = sm_get_handle(fd)))
  {
    ret = sm_uscsi_cmd(sm_hdl, psc);
    if (ret < 0)
    {
      ret = ioctl (fd, USCSICMD, psc);
    }
    sm_release_handle(sm_hdl);
  }
  return ret;
}

#ifdef __cplusplus
}
#endif

// vim: ts=2 sw=2 et

